from __future__ import print_function

import renpy
renpy.update_path()

from renpy.uguu.uguu import *


# Generated by scripts/relative_imports.py, do not edit below this line.
if 1 == 0:
    from . import gl
    from . import uguu
