/* fribidi-unicode-version.h
 * generated by gen-unicode-version (GNU FriBidi 0.19.1)
 * from the file BidiMirroring.txt */

#define FRIBIDI_UNICODE_VERSION "5.1.0"
#define FRIBIDI_UNICODE_MAJOR_VERSION 5
#define FRIBIDI_UNICODE_MINOR_VERSION 1
#define FRIBIDI_UNICODE_MICRO_VERSION 0
#define FRIBIDI_UNICODE_NAMESPACE(SYMBOL) \
	FRIBIDI_NAMESPACE(SYMBOL##_unicode_5_1_0)
#define FRIBIDI_UNICODE_PRIVATESPACE(SYMBOL) \
	FRIBIDI_PRIVATESPACE(SYMBOL##_unicode_5_1_0)

/* End of generated fribidi-unicode-version.h */
