from __future__ import print_function

import renpy
renpy.update_path()


# Generated by scripts/relative_imports.py, do not edit below this line.
if 1 == 0:
    from . import gl2debug
    from . import gl2draw
    from . import gl2functions
    from . import gl2mesh
    from . import gl2mesh2
    from . import gl2mesh3
    from . import gl2model
    from . import gl2polygon
    from . import gl2shader
    from . import gl2shadercache
    from . import gl2texture
    from . import live2d
    from . import live2dmodel
    from . import live2dmotion
