/* lib/fribidi-config.h.  Generated from fribidi-config.h.in by configure. */
/* Not copyrighted, in public domain. */
#ifndef FRIBIDI_CONFIG_H
#define FRIBIDI_CONFIG_H

#define FRIBIDI "fribidi"
#define FRIBIDI_NAME "GNU FriBidi"
#define FRIBIDI_BUGREPORT "http://fribidi.org/bug"

#define FRIBIDI_VERSION "0.19.2"
#define FRIBIDI_MAJOR_VERSION 0
#define FRIBIDI_MINOR_VERSION 19
#define FRIBIDI_MICRO_VERSION 2
#define FRIBIDI_INTERFACE_VERSION 3
#define FRIBIDI_INTERFACE_VERSION_STRING "3"

/* Define to 1 if you want charset conversion codes in the library */
#define FRIBIDI_CHARSETS 0

/* Define to 1 if you want to use glib */
#define FRIBIDI_USE_GLIB 0

/* The size of a `int', as computed by sizeof. */
#define FRIBIDI_SIZEOF_INT 4

#endif /* FRIBIDI_CONFIG_H */
